// -*- Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*-
//
// Copyright (c) 2006 Georgia Tech Research Corporation
// Copyright (c) 2010 INRIA
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Author: George F. Riley<riley@ece.gatech.edu>
// Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
//

#include "ns3/packet.h"
#include "ns3/log.h"
#include "ns3/callback.h"
#include "ns3/ipv4-address.h"
#include "ns3/ipv4-route.h"
#include "ns3/node.h"
#include "ns3/socket.h"
#include "ns3/net-device.h"
#include "ns3/uinteger.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/object-vector.h"
#include "ns3/ipv4-header.h"
#include "ns3/boolean.h"
#include "ns3/ipv4-routing-table-entry.h"

#include "loopback-net-device.h"
#include "arp-l3-protocol.h"
#include "ipv4-l3-protocol.h"
#include "ipv4-l4-protocol.h"
#include "icmpv4-l4-protocol.h"
#include "ipv4-interface.h"
#include "ipv4-raw-socket-impl.h"
#include "ns3/wifi-net-device.h"
#include "ns3/config.h"
#include "ns3/address-utils.h"
#include "ns3/simulator.h"
#include "ns3/ipv4-list-routing.h"
#include "ns3/ipv4-static-routing.h"

/////////////////Naveed//////////////////
#define  BUFFER
//#define  DEBUG
#define  MAX_NUM_FRAMES_TO_SEND  10
#define  MAX_NUM_OF_COPIES   10


#include "ns3/udp-header.h"
#include "ns3/seq-header.h"
/////////////////Naveed//////////////////

NS_LOG_COMPONENT_DEFINE ("Ipv4L3Protocol");

namespace ns3 {

const uint16_t Ipv4L3Protocol::PROT_NUMBER = 0x0800;

NS_OBJECT_ENSURE_REGISTERED (Ipv4L3Protocol);

TypeId 
Ipv4L3Protocol::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Ipv4L3Protocol")
    .SetParent<Ipv4> ()
    .AddConstructor<Ipv4L3Protocol> ()
    .AddAttribute ("DefaultTtl", "The TTL value set by default on all outgoing packets generated on this node.",
                   UintegerValue (64),
                   MakeUintegerAccessor (&Ipv4L3Protocol::m_defaultTtl),
                   MakeUintegerChecker<uint8_t> ())
    .AddTraceSource ("Tx", "Send ipv4 packet to outgoing interface.",
                   MakeTraceSourceAccessor (&Ipv4L3Protocol::m_txTrace))
    .AddTraceSource ("Rx", "Receive ipv4 packet from incoming interface.",
                     MakeTraceSourceAccessor (&Ipv4L3Protocol::m_rxTrace))
    .AddTraceSource ("Drop", "Drop ipv4 packet",
                     MakeTraceSourceAccessor (&Ipv4L3Protocol::m_dropTrace))
    .AddAttribute ("InterfaceList", "The set of Ipv4 interfaces associated to this Ipv4 stack.",
                   ObjectVectorValue (),
                   MakeObjectVectorAccessor (&Ipv4L3Protocol::m_interfaces),
                   MakeObjectVectorChecker<Ipv4Interface> ())
       
    .AddTraceSource ("NotifyAssoc", "Notify about remote association.",
                   MakeTraceSourceAccessor (&Ipv4L3Protocol::m_notifyRemoteAssociation))
    ;
  return tid;
}

Ipv4L3Protocol::Ipv4L3Protocol()
  : m_nInterfaces (0),
    m_identification (0)
{
  NS_LOG_FUNCTION_NOARGS ();
  
  ////////////////Naveed/////////////////
  CreateApIpv4List ();
  m_txTimer = Seconds (0.05);
  ////////////////Naveed/////////////////
}

Ipv4L3Protocol::~Ipv4L3Protocol ()
{
  NS_LOG_FUNCTION (this);
  if (m_txTimerEvent.IsRunning ())
    {
      m_txTimerEvent.Cancel ();
    }
}

void
Ipv4L3Protocol::Insert(Ptr<Ipv4L4Protocol> protocol)
{
  m_protocols.push_back (protocol);
}
Ptr<Ipv4L4Protocol>
Ipv4L3Protocol::GetProtocol(int protocolNumber) const
{
  for (L4List_t::const_iterator i = m_protocols.begin(); i != m_protocols.end(); ++i)
    {
      if ((*i)->GetProtocolNumber () == protocolNumber)
  {
    return *i;
  }
    }
  return 0;
}
void
Ipv4L3Protocol::Remove (Ptr<Ipv4L4Protocol> protocol)
{
  m_protocols.remove (protocol);
}

void
Ipv4L3Protocol::SetNode (Ptr<Node> node)
{
  m_node = node;
  // Add a LoopbackNetDevice if needed, and an Ipv4Interface on top of it
  SetupLoopback ();
  
  //////////////////////Naveed////////////////////
  ConnectToCatchAssocEvent ();
  ////////////////////////////////////////////////
}

Ptr<Socket> 
Ipv4L3Protocol::CreateRawSocket (void)
{
  NS_LOG_FUNCTION (this);
  Ptr<Ipv4RawSocketImpl> socket = CreateObject<Ipv4RawSocketImpl> ();
  socket->SetNode (m_node);
  m_sockets.push_back (socket);
  return socket;
}
void 
Ipv4L3Protocol::DeleteRawSocket (Ptr<Socket> socket)
{
  NS_LOG_FUNCTION (this << socket);
  for (SocketList::iterator i = m_sockets.begin (); i != m_sockets.end (); ++i)
    {
      if ((*i) == socket)
        {
          m_sockets.erase (i);
          return;
        }
    }
  return;
}
/*
 * This method is called by AddAgregate and completes the aggregation
 * by setting the node in the ipv4 stack
 */
void
Ipv4L3Protocol::NotifyNewAggregate ()
{
  Ptr<Node>node = this->GetObject<Node>();
  // verify that it's a valid node and that
  // the node has not been set before
  if (node!= 0 && m_node == 0)
    {
      this->SetNode (node);
    }
  Object::NotifyNewAggregate ();
}

void 
Ipv4L3Protocol::SetRoutingProtocol (Ptr<Ipv4RoutingProtocol> routingProtocol)
{
  NS_LOG_FUNCTION (this);
  m_routingProtocol = routingProtocol;
  m_routingProtocol->SetIpv4 (this);
}


Ptr<Ipv4RoutingProtocol> 
Ipv4L3Protocol::GetRoutingProtocol (void) const
{
  return m_routingProtocol;
}

void 
Ipv4L3Protocol::DoDispose (void)
{
  NS_LOG_FUNCTION (this);
  for (L4List_t::iterator i = m_protocols.begin(); i != m_protocols.end(); ++i)
    {
      *i = 0;
    }
  m_protocols.clear ();

  for (Ipv4InterfaceList::iterator i = m_interfaces.begin (); i != m_interfaces.end (); ++i)
    {
      *i = 0;
    }
  m_interfaces.clear ();
  m_node = 0;
  m_routingProtocol = 0;
  Object::DoDispose ();
}

void
Ipv4L3Protocol::SetupLoopback (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  Ptr<Ipv4Interface> interface = CreateObject<Ipv4Interface> ();
  Ptr<LoopbackNetDevice> device = 0;
  // First check whether an existing LoopbackNetDevice exists on the node
  for (uint32_t i = 0; i < m_node->GetNDevices (); i++)
    {
      if (device = DynamicCast<LoopbackNetDevice> (m_node->GetDevice (i)))
        {
          break;
        }
    }
  if (device == 0)
    {
      device = CreateObject<LoopbackNetDevice> (); 
      m_node->AddDevice (device);
    }
  interface->SetDevice (device);
  interface->SetNode (m_node);
  Ipv4InterfaceAddress ifaceAddr = Ipv4InterfaceAddress (Ipv4Address::GetLoopback (), Ipv4Mask::GetLoopback ());
  interface->AddAddress (ifaceAddr);
  uint32_t index = AddIpv4Interface (interface);
  Ptr<Node> node = GetObject<Node> ();
  node->RegisterProtocolHandler (MakeCallback (&Ipv4L3Protocol::Receive, this), 
                                 Ipv4L3Protocol::PROT_NUMBER, device);
  interface->SetUp ();
  if (m_routingProtocol != 0)
    {
      m_routingProtocol->NotifyInterfaceUp (index);
    }
}

void 
Ipv4L3Protocol::SetDefaultTtl (uint8_t ttl)
{
  NS_LOG_FUNCTION_NOARGS ();
  m_defaultTtl = ttl;
}
    
uint32_t 
Ipv4L3Protocol::AddInterface (Ptr<NetDevice> device)
{
  NS_LOG_FUNCTION (this << &device);

  Ptr<Node> node = GetObject<Node> ();
  node->RegisterProtocolHandler (MakeCallback (&Ipv4L3Protocol::Receive, this), 
                                 Ipv4L3Protocol::PROT_NUMBER, device);
  node->RegisterProtocolHandler (MakeCallback (&ArpL3Protocol::Receive, PeekPointer (GetObject<ArpL3Protocol> ())),
                                 ArpL3Protocol::PROT_NUMBER, device);

  Ptr<Ipv4Interface> interface = CreateObject<Ipv4Interface> ();
  interface->SetNode (m_node);
  interface->SetDevice (device);
  interface->SetForwarding (m_ipForward);
  return AddIpv4Interface (interface);
}

uint32_t 
Ipv4L3Protocol::AddIpv4Interface (Ptr<Ipv4Interface>interface)
{
  NS_LOG_FUNCTION (this << interface);
  uint32_t index = m_nInterfaces;
  m_interfaces.push_back (interface);
  m_nInterfaces++;
  return index;
}

Ptr<Ipv4Interface>
Ipv4L3Protocol::GetInterface (uint32_t index) const
{
  NS_LOG_FUNCTION (this << index);
  uint32_t tmp = 0;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); i != m_interfaces.end (); i++)
    {
      if (index == tmp) 
  {
    return *i;
  }
      tmp++;
    }
  return 0;
}

uint32_t 
Ipv4L3Protocol::GetNInterfaces (void) const
{
  NS_LOG_FUNCTION_NOARGS ();
  return m_nInterfaces;
}

int32_t 
Ipv4L3Protocol::GetInterfaceForAddress (
  Ipv4Address address) const
{
  NS_LOG_FUNCTION (this << address);

  int32_t interface = 0;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); 
       i != m_interfaces.end (); 
       i++, interface++)
    {
      for (uint32_t j = 0; j < (*i)->GetNAddresses (); j++)
        {
          if ((*i)->GetAddress (j).GetLocal () == address)
            {
              return interface;
            }
        }
    }

  return -1;
}

int32_t 
Ipv4L3Protocol::GetInterfaceForPrefix (
  Ipv4Address address, 
  Ipv4Mask mask) const
{
  NS_LOG_FUNCTION (this << address << mask);

  int32_t interface = 0;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); 
       i != m_interfaces.end (); 
       i++, interface++)
    {
      for (uint32_t j = 0; j < (*i)->GetNAddresses (); j++)
        {
          if ((*i)->GetAddress (j).GetLocal ().CombineMask (mask) == address.CombineMask (mask))
            {
              return interface;
            }
        }
    }

  return -1;
}

int32_t 
Ipv4L3Protocol::GetInterfaceForDevice (
  Ptr<const NetDevice> device) const
{
  NS_LOG_FUNCTION (this << device->GetIfIndex());

  int32_t interface = 0;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); 
       i != m_interfaces.end (); 
       i++, interface++)
    {
      if ((*i)->GetDevice () == device)
        {
          return interface;
        }
    }

  return -1;
}

void 
Ipv4L3Protocol::Receive( Ptr<NetDevice> device, Ptr<const Packet> p, uint16_t protocol, const Address &from,
                         const Address &to, NetDevice::PacketType packetType)
{
  NS_LOG_FUNCTION (this << &device << p << protocol <<  from);

  NS_LOG_LOGIC ("Packet from " << from << " received on node " << 
    m_node->GetId ());

  uint32_t interface = 0;
  Ptr<Packet> packet = p->Copy ();

  Ptr<Ipv4Interface> ipv4Interface;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); 
       i != m_interfaces.end (); 
       i++)
    {
      ipv4Interface = *i;
      if (ipv4Interface->GetDevice () == device)
        {
          if (ipv4Interface->IsUp ())
            {
              m_rxTrace (packet, interface);
              break;
            }
          else
            {
              NS_LOG_LOGIC ("Dropping received packet-- interface is down");
              m_dropTrace (packet);
              return;
            }
        }
      interface++;
    }

  Ipv4Header ipHeader;
  if (Node::ChecksumEnabled ())
    {
      ipHeader.EnableChecksum ();
    }
  packet->RemoveHeader (ipHeader);

  if (!ipHeader.IsChecksumOk ()) 
    {
      m_dropTrace (packet);
      return;
    }
    
  
  ////////////////////Naveed//////////////////////  
  bool flagIP = false;
  for (uint32_t j = 0; j < GetNInterfaces (); j++)
    {
      if (GetAddress (j, 0).GetLocal () == ipHeader.GetDestination ())
        {
          flagIP = true;
          break;
        }
    }    
    
  if (ipHeader.GetProtocol () == 18 && flagIP == true)
    {
      std::cout << "Check Point 1: Packet Received for a Manet Node with IP Encapsulation\n";
      MedehaIntraManetNotifHeader hdr;
      packet->RemoveHeader (hdr);
      
      ipHeader.SetDestination (hdr.GetDstIPAddress ());
      ipHeader.SetProtocol (hdr.GetProtocol ());
    }


  Ipv4Address ipAddress ("127.0.0.1");
  bool found = false;
  for (uint32_t j = 0; j < ipv4Interface->GetNAddresses (); j++)
    {
      if (ipv4Interface->GetAddress (j).GetLocal () == ipAddress)
        {
          found = true;
          break;
        }
    }
  if (ipHeader.GetSource () == "10.1.1.1" ||
      ipHeader.GetSource () == "10.1.1.2" ||
      ipHeader.GetSource () == "10.1.1.3" ||
      ipHeader.GetSource () == "10.1.1.4" ||
      ipHeader.GetSource () == "10.1.1.5" ||
      ipHeader.GetSource () == "10.1.1.6" ||
      ipHeader.GetSource () == "10.1.1.7" ||
      ipHeader.GetSource () == "10.1.1.8" ||
      ipHeader.GetSource () == "10.1.1.9" ||
      ipHeader.GetDestination () == "10.1.1.1" ||
      ipHeader.GetDestination () == "10.1.1.2" ||
      ipHeader.GetDestination () == "10.1.1.3" ||
      ipHeader.GetDestination () == "10.1.1.4" ||
      ipHeader.GetDestination () == "10.1.1.5" ||
      ipHeader.GetDestination () == "10.1.1.6" ||
      ipHeader.GetDestination () == "10.1.1.7" ||
      ipHeader.GetDestination () == "10.1.1.8" ||
      ipHeader.GetDestination () == "10.1.1.9")
    {
      if (!found)
        {
          Ipv4Address addr;
          for (uint32_t i = 0; i < GetNInterfaces (); i++)
            {
              addr = GetAddress (i, 0).GetLocal ();
            }
                 
          APNotifL3Header apHdr;
          packet->RemoveHeader (apHdr);
          int status = m_hNotif.ParseNotification (packet, apHdr);
          if (status)
            {
              uint32_t ifIndex = GetInterfaceForDevice (device);
              Ptr<Ipv4ListRouting> ipv4LRP;
              ipv4LRP = DynamicCast<Ipv4ListRouting> (m_routingProtocol);
          
              int16_t priority1 = 0;
              Ptr<Ipv4StaticRouting> ipv4Static =  DynamicCast<Ipv4StaticRouting> (ipv4LRP->GetRoutingProtocol (0, priority1));
    
              Ipv4Address* addresses = apHdr.GetIPAddresses ();
              if (status == ASS_NOTIF)
                {
                  RemoveRoutingEntry (addresses[0], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[0], apHdr.GetSrcIPAddress (), ifIndex);
            
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      RemoveRoutingEntry (addresses[1], ifIndex);
                      ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
                    }
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      RemoveRoutingEntry (addresses[2], ifIndex);
                      ipv4Static->AddHostRouteTo (addresses[2], apHdr.GetSrcIPAddress (), ifIndex);
                    }
        
                  Ipv4Address assAPAddress, infraStructureAddress;
                  int status = m_hNotif.CheckForDestInfo (addresses[0], &assAPAddress, &infraStructureAddress);
        
                  if (status == DEST_INFO_ENTRY_FOUND ||
                      status == DEST_INFO_ALTERNATE_FOUND)
                    {
                      uint16_t len;
                      m_hNotif.RemoveDestInfo (addresses[0], &infraStructureAddress, &len);
                      m_hNotif.RemoveDestBufferInfo (addresses[0]);
                    }
    
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[1], &assAPAddress, &infraStructureAddress);
        
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[1], &infraStructureAddress, &len);
                          m_hNotif.RemoveDestBufferInfo (addresses[1]);
                        }
                    }
    
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[2], &assAPAddress, &infraStructureAddress);
        
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[2], &infraStructureAddress, &len);
                          m_hNotif.RemoveDestBufferInfo (addresses[2]);
                        }   
                    }
    
                  m_hNotif.AddOrModifyDestInfo(addresses, apHdr.GetSrcIPAddress (), apHdr.GetNumOfIfces (), apHdr.GetManetConnection ());
/*
              RemoveRoutingEntry (addresses[0], ifIndex);
              ipv4Static->AddHostRouteTo (addresses[0], apHdr.GetSrcIPAddress (), ifIndex);
        
        if (apHdr.GetNumOfIfces () >= 2)
          {
      RemoveRoutingEntry (addresses[1], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
    }
              if (apHdr.GetNumOfIfces () >= 3)
          {
      RemoveRoutingEntry (addresses[2], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[2], apHdr.GetSrcIPAddress (), ifIndex);
    }
*/
#ifdef DEBUG
                  std::cout << ipHeader.GetDestination () << ", IPv4L3Protocol:Receive: Adding route for address " << addresses[0] << "\n";
                  std::cout << ipHeader.GetDestination () << ", IPv4L3Protocol:Receive: Adding route for address " << addresses[1] << "\n";
#endif
/*       
        Ptr<ArpL3Protocol> arp = m_node->GetObject<ArpL3Protocol> ();
              Address addressSta = arp->SearchForMacAddress (apHdr.GetIPAddress01 ());
        Mac48Address macAddressSta;
              macAddressSta = Mac48Address::ConvertFrom (addressSta);
        m_notifyRemoteAssociation (macAddressSta);*/
                } 
              else if (status == DISASS_NOTIF)
                {
//        uint32_t ifIndex = GetInterfaceForDevice (device);
//        RemoveRoutingEntry (apHdr.GetSrcIPAddress (), ifIndex);
/*        if (buffer.m_listSize == 0)
    {
      if (m_txTimerEvent.IsRunning ())
          m_txTimerEvent.Cancel ();
    }*/
                }
              else if (status == FETCH_FRAMES_NOTIF)
                {
        // count the number of stored frames for the destination
                  int numStoredFrames01 = CountNumberOfStoredFrames (addresses[0]);
                  int numStoredFrames02 = CountNumberOfStoredFrames (addresses[1]);
                  int numStoredFrames03 = CountNumberOfStoredFrames (addresses[2]);
#ifdef DEBUG
                  std::cout << "Number of Stored Frames for dest = " << addresses[0] << " are " << numStoredFrames01 << "\n";
                  std::cout << "Number of Stored Frames for dest = " << addresses[1] << " are " << numStoredFrames02 << "\n";
#endif
    
              // check if there are still some frames to be sent for the destination "tempAddress"
                  if (numStoredFrames01 > 0 ||
                      numStoredFrames02 > 0 ||
                      numStoredFrames03 > 0)
                    {
                      if (m_txTimerEvent.IsRunning ())
                          m_txTimerEvent.Cancel ();
                  // send frames to the corresponding AP
                      SendStoreFramesToDest();
                      m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
                    }
                }
              else if (status == NEIGHBORHOOD_INFO_NOTIF)
                {
#ifdef DEBUG
                  std::cout << "Sender Address = " << apHdr.GetSrcIPAddress () << " Address 1 = " << addresses[0] << " Address 2 = " << addresses[1] << "\n";
#endif            
        
        // First check if an entry to association info list is already present
        // which would mean that the station is already directly associated with the AP
        // and hence there is no need for an indirect path to destination
                  Ipv4Address assAPAddress, infraStructureAddress;
              
        // these variables are initialed to facilitate the AND operation in the 'if' statement below
        // as if there are only one or two interfaces, the AND operation should have the possibility to be true
                  int status0 = DEST_INFO_ENTRY_NOT_FOUND;
                  int status1 = DEST_INFO_ENTRY_NOT_FOUND;
                  int status2 = DEST_INFO_ENTRY_NOT_FOUND;
                  
                  status0 = m_hNotif.CheckForDestInfo (addresses[0], &assAPAddress, &infraStructureAddress);
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      status1 = m_hNotif.CheckForDestInfo (addresses[1], &assAPAddress, &infraStructureAddress);
                    }
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      status2 = m_hNotif.CheckForDestInfo (addresses[2], &assAPAddress, &infraStructureAddress);
                    }
                  
                  if (status0 == DEST_INFO_LIST_DOESNT_EXIST ||
                      status1 == DEST_INFO_LIST_DOESNT_EXIST ||
                      status2 == DEST_INFO_LIST_DOESNT_EXIST ||
                      (status0 == DEST_INFO_ENTRY_NOT_FOUND &&
                       status1 == DEST_INFO_ENTRY_NOT_FOUND &&
                       status2 == DEST_INFO_ENTRY_NOT_FOUND))
                    {
                      m_hNotif.RemoveDestBufferInfo (addresses[0]);     
                      m_hNotif.AddOrModifyDestInfo (addresses, apHdr.GetSrcIPAddress (), apHdr.GetNumOfIfces (), apHdr.GetManetConnection ());
      
                      RemoveRoutingEntry (addresses[0], ifIndex);
                      ipv4Static->AddHostRouteTo (addresses[0], apHdr.GetSrcIPAddress (), ifIndex);
      
                      if (apHdr.GetNumOfIfces () >= 2)
                        {
                          m_hNotif.RemoveDestBufferInfo (addresses[1]);
                          RemoveRoutingEntry (addresses[1], ifIndex);
                          ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
                        }
                      if (apHdr.GetNumOfIfces () >= 3)
                        {
                          m_hNotif.RemoveDestBufferInfo (addresses[2]);
                          RemoveRoutingEntry (addresses[2], ifIndex);
                          ipv4Static->AddHostRouteTo (addresses[2], apHdr.GetSrcIPAddress (), ifIndex);
                        }
        
                      Ptr<Packet> newPacket = Create<Packet> ();
                      m_hNotif.SendNotificationToAPs (INDIRECT_ASSOCIATION_NOTIF, apHdr.GetNumOfIfces (), ipHeader.GetDestination (), addresses, 0, newPacket);
                      
                      ApIpv4AddressList::iterator newIterator;
                      Ipv4Address destination;
                      for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
                        {
                          Ptr<Packet> p = newPacket->Copy ();
                          destination = *newIterator;

                          if (CheckForLocalInterface (destination))
                            {
                              continue;
                            }
                          else
                            {
                              Send (p, ipHeader.GetDestination (), destination, 17, NULL);
                            }
                        }
/*        
      Ptr<Packet> packet2 = Create<Packet> ();
                  m_hNotif.SendNotificationToAPs (FETCH_FRAMES_NOTIF, 0, ipHeader.GetDestination (), apHdr.GetIPAddress01 (), apHdr.GetIPAddress02 (), packet2);
  
                  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
                    {
                      Ptr<Packet> p = packet2->Copy ();
                      destination = *newIterator;

                      if (CheckForLocalInterface (destination))
                        {
                          continue;
                        }
                      else
                        {
                          Send (p, ipHeader.GetDestination (), destination, 17, NULL);
                  }
                    }
    
                  std::cout << "Ipv4L3Protocol::Receive (), Number of Stored Frames for dest " << apHdr.GetIPAddress01 () << " is = " << numStoredFrames01 << "\n";
                  std::cout << "Ipv4L3Protocol::Receive (), Number of Stored Frames for dest " << apHdr.GetIPAddress02 () << " is = " << numStoredFrames02 << "\n";       
                  std::cout << "Ipv4L3Protocol::Receive (), Total Stored Frames are = " << buffer.m_listSize << "\n";                  
*/
/*
            if (apHdr.GetNumOfIfces () == 2)
              {
                RemoveRoutingEntry (addresses[1], ifIndex);
                ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
      
          Ptr<Packet> newPacket2 = Create<Packet> ();
////////////////      m_hNotif.SendNotificationToAPs (INDIRECT_ASSOCIATION_NOTIF, 0, ipHeader.GetDestination (), apHdr.GetIPAddress02 (), ipHeader.GetDestination (), newPacket2);
        
                for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
                        {
                          Ptr<Packet> p = newPacket2->Copy ();
                          destination = *newIterator;

                          if (CheckForLocalInterface (destination))
                            {
                        continue;
                      }
                          else
                            {
                              Send (p, ipHeader.GetDestination (), destination, 17, NULL);
                      }
                        }
        }
*/
                      int numStoredFrames01 = CountNumberOfStoredFrames (addresses[0]);
                      int numStoredFrames02 = CountNumberOfStoredFrames (addresses[1]);
                      int numStoredFrames03 = CountNumberOfStoredFrames (addresses[2]);

#ifdef DEBUG      
                      std::cout << "Ipv4L3Protocol::Receive (), Number of Stored Frames for dest " << addresses[0] << " and for dest " << addresses[1] << " are " << numStoredFrames01 << ":" << numStoredFrames02 << std::endl;
#endif
      
                      if (numStoredFrames01 > 0 ||
                          numStoredFrames02 > 0 ||
                          numStoredFrames03 > 0)
                        {
                          if (m_txTimerEvent.IsRunning ())
                              m_txTimerEvent.Cancel ();
                          SendStoreFramesToDest ();
                          m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
                        }
                    }
                  else
                    {
#ifdef DEBUG
                      std::cout << "Destination " << addresses[0] << " is already associated with the AP, Ignoring Neighborhood Info\n";
#endif
                    }
                }
              else if (status == INDIRECT_ASSOCIATION_NOTIF)
                {
                  RemoveRoutingEntry (addresses[0], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[0], apHdr.GetSrcIPAddress (), ifIndex);
                  m_hNotif.RemoveDestBufferInfo (addresses[0]);
        
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      RemoveRoutingEntry (addresses[1], ifIndex);
                      ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
                      m_hNotif.RemoveDestBufferInfo (addresses[1]);
                    }
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      RemoveRoutingEntry (addresses[2], ifIndex);
                      ipv4Static->AddHostRouteTo (addresses[2], apHdr.GetSrcIPAddress (), ifIndex);
                      m_hNotif.RemoveDestBufferInfo (addresses[2]);
                    }
        
                  Ipv4Address assAPAddress, infraStructureAddress;
                  int status = m_hNotif.CheckForDestInfo (addresses[0], &assAPAddress, &infraStructureAddress);
                  
                  if (status == DEST_INFO_ENTRY_FOUND ||
                      status == DEST_INFO_ALTERNATE_FOUND)
                    {
                      uint16_t len;
                      m_hNotif.RemoveDestInfo (addresses[0], &infraStructureAddress, &len);
                    }
    
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[1], &assAPAddress, &infraStructureAddress);
        
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[1], &infraStructureAddress, &len);
                        }
                    }
    
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[2], &assAPAddress, &infraStructureAddress);
            
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[2], &infraStructureAddress, &len);
                        }   
                    }
    
                  m_hNotif.AddOrModifyDestInfo(addresses, apHdr.GetSrcIPAddress (), apHdr.GetNumOfIfces (), apHdr.GetManetConnection ());
        
/*              
              RemoveRoutingEntry (addresses[0], ifIndex);
              ipv4Static->AddHostRouteTo (addresses[0], apHdr.GetSrcIPAddress (), ifIndex);
        
        if (apHdr.GetNumOfIfces () >= 2)
          {
                  RemoveRoutingEntry (addresses[1], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[1], apHdr.GetSrcIPAddress (), ifIndex);
                }
        if (apHdr.GetNumOfIfces () >= 3)
          {
                  RemoveRoutingEntry (addresses[2], ifIndex);
                  ipv4Static->AddHostRouteTo (addresses[2], apHdr.GetSrcIPAddress (), ifIndex);
                }
*/    
                  int numStoredFrames01 = CountNumberOfStoredFrames (addresses[0]);
                  int numStoredFrames02 = CountNumberOfStoredFrames (addresses[1]);
                  int numStoredFrames03 = CountNumberOfStoredFrames (addresses[2]);

#ifdef DEBUG
                  std::cout << "Ipv4L3Protocol::Receive (), Number of Stored Frames for dest " << addresses[0] << " is = " << numStoredFrames01 << "\n";
                  std::cout << "Ipv4L3Protocol::Receive (), Number of Stored Frames for dest " << addresses[1] << " is = " << numStoredFrames02 << "\n";        
                  std::cout << "Ipv4L3Protocol::Receive (), Total Stored Frames are = " << buffer.m_listSize << "\n";
#endif

                  if (numStoredFrames01 > 0 ||
                      numStoredFrames02 > 0 ||
                      numStoredFrames03 > 0)
                    {
                      if (m_txTimerEvent.IsRunning ())
                          m_txTimerEvent.Cancel ();
                      SendStoreFramesToDest();
                      m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
                    }
                }
              else if (status == STA_ASS_NOTIF)
                {
                  RemoveRoutingEntry (addresses[0], ifIndex);
        
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      RemoveRoutingEntry (addresses[1], ifIndex);
                    }
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      RemoveRoutingEntry (addresses[2], ifIndex);
                    }
        
                  Ipv4Address assAPAddress, infraStructureAddress;
                  int status = m_hNotif.CheckForDestInfo (addresses[0], &assAPAddress, &infraStructureAddress);
        
                  if (status == DEST_INFO_ENTRY_FOUND ||
                  status == DEST_INFO_ALTERNATE_FOUND)
                    {
                      uint16_t len;
                      m_hNotif.RemoveDestInfo (addresses[0], &infraStructureAddress, &len);
                    }
    
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[1], &assAPAddress, &infraStructureAddress);
        
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[1], &infraStructureAddress, &len);
                        }
                    }
    
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      status = m_hNotif.CheckForDestInfo (addresses[2], &assAPAddress, &infraStructureAddress);
             
                      if (status == DEST_INFO_ENTRY_FOUND ||
                          status == DEST_INFO_ALTERNATE_FOUND)
                        {
                          uint16_t len;
                          m_hNotif.RemoveDestInfo (addresses[2], &infraStructureAddress, &len);
                        }   
                    }

                  Ptr<Packet> packet1 = Create<Packet> ();
                  m_hNotif.AddOrModifyDestInfo (addresses, apHdr.GetSrcIPAddress (), apHdr.GetNumOfIfces (), apHdr.GetManetConnection ());
                  m_hNotif.SendNotificationToAPs (ASS_NOTIF, apHdr.GetNumOfIfces (), apHdr.GetSrcIPAddress (), addresses, 0, packet1);
 
                  ApIpv4AddressList::iterator newIterator;
                  Ipv4Address destination;
  
              //  std::cout << "ReceiveAssocNotif: Adding route for address " << ipv4AddressSta << "\n";
/*
              Ptr<Ipv4ListRouting> ipv4LRP;
              ipv4LRP = DynamicCast<Ipv4ListRouting> (m_routingProtocol);
              int16_t priority1 = 0;
              Ptr<Ipv4StaticRouting> ipv4Static =  DynamicCast<Ipv4StaticRouting> (ipv4LRP->GetRoutingProtocol (0, priority1));
*/
                  ipv4Static->AddHostRouteTo (addresses[0], ifIndex);
        
                  if (apHdr.GetNumOfIfces () >= 2)
                    {
                      ipv4Static->AddHostRouteTo (addresses[1], addresses[0], ifIndex);
                    }
                  if (apHdr.GetNumOfIfces () >= 3)
                    {
                      ipv4Static->AddHostRouteTo (addresses[2], addresses[0], ifIndex);
                    }
            
                  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
                    {
                      Ptr<Packet> p = packet1->Copy ();
                      destination = *newIterator;
                      if (CheckForLocalInterface (destination))
                        {
                          continue;
                        }
                      else
                        {
                          Send (p, ipHeader.GetDestination (), destination, 17, NULL);
                        }
                    }
                  m_hNotif.RemoveDestBufferInfo (addresses[0]);
        
                  Ptr<Packet> packet2 = Create<Packet> ();
                  m_hNotif.SendNotificationToAPs (FETCH_FRAMES_NOTIF, apHdr.GetNumOfIfces (), apHdr.GetSrcIPAddress (), addresses, 0, packet2);
      
                  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
                    {
                      Ptr<Packet> p = packet2->Copy ();
                      destination = *newIterator;
     
                      if (CheckForLocalInterface (destination))
                        {
                          continue;
                        }
                      else
                        {
                          Send (p, ipHeader.GetDestination (), destination, 17, NULL);
                        }
                    }
    
                  int numStoredFrames01 = CountNumberOfStoredFrames(addresses[0]);
                  int numStoredFrames02 = CountNumberOfStoredFrames(addresses[1]);
                  int numStoredFrames03 = CountNumberOfStoredFrames(addresses[2]);

#ifdef DEBUG
                  std::cout << "ReceiveAssocNotif (), Number of Stored Frames for dest " << addresses[0] << " is = " << numStoredFrames01 << "\n";
                  std::cout << "ReceiveAssocNotif (), Number of Stored Frames for dest " << addresses[1] << " is = " << numStoredFrames02 << "\n";        
                  std::cout << "ReceiveAssocNotif (), Total Stored Frames are = " << buffer.m_listSize << "\n";
#endif

                  if (numStoredFrames01 > 0 ||
                      numStoredFrames02 > 0 ||
                      numStoredFrames03 > 0)
                    {
                      if (m_txTimerEvent.IsRunning ())
                          m_txTimerEvent.Cancel ();
                      SendStoreFramesToDest();
                      m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
                    }
                }
      
              else if (status == NEIGHBORHOOD_DISASS_NOTIF)
                {
                  Ipv4Address alternateAddress;
                  Ipv4Address apAddress;
                  m_hNotif.CheckForDestInfo (addresses[0], &apAddress, &alternateAddress);

                  if (apAddress == apHdr.GetSrcIPAddress ())
                    {
#ifdef DEBUG
                      std::cout << "Ipv4L3Protocol::Receive (), Removing entry for indirect association\n";
#endif
                      uint16_t len;
                      status = m_hNotif.RemoveDestInfo (addresses[0], &alternateAddress, &len);
                      status = m_hNotif.CheckAndAddDestBufferAP (addresses, ipHeader.GetDestination (), apHdr.GetNumOfIfces ());
                      RemoveRoutingEntry (addresses[0], ifIndex);
                      if (apHdr.GetNumOfIfces () >= 2)
                        {
                          RemoveRoutingEntry (addresses[1], ifIndex);
                        }
                      if (apHdr.GetNumOfIfces () >= 3)
                        {
                          RemoveRoutingEntry (addresses[2], ifIndex);
                        }
                    }
                }
    
              else if (status == MANET_DISCONNECT_NOTIF)
                {
                  RemoveManetRoutes (apHdr.GetSrcIPAddress (), ipHeader.GetDestination ());
                }
    
              return;
            } 
        }
    }
  ////////////////////Naveed//////////////////////

  for (SocketList::iterator i = m_sockets.begin (); i != m_sockets.end (); ++i)
    {
      Ptr<Ipv4RawSocketImpl> socket = *i;
      socket->ForwardUp (packet, ipHeader, device);
    }

  m_routingProtocol->RouteInput (packet, ipHeader, device, 
    MakeCallback (&Ipv4L3Protocol::IpForward, this),
    MakeCallback (&Ipv4L3Protocol::IpMulticastForward, this),
    MakeCallback (&Ipv4L3Protocol::LocalDeliver, this),
    MakeCallback (&Ipv4L3Protocol::RouteInputError, this)
  );

}

Ptr<Icmpv4L4Protocol> 
Ipv4L3Protocol::GetIcmp (void) const
{
  Ptr<Ipv4L4Protocol> prot = GetProtocol (Icmpv4L4Protocol::GetStaticProtocolNumber ());
  if (prot != 0)
    {
      return prot->GetObject<Icmpv4L4Protocol> ();
    }
  else
    {
      return 0;
    }
}

bool
Ipv4L3Protocol::IsUnicast (Ipv4Address ad, Ipv4Mask interfaceMask) const
{
  return !ad.IsMulticast () && !ad.IsSubnetDirectedBroadcast (interfaceMask);
}

void 
Ipv4L3Protocol::Send (Ptr<Packet> packet, 
            Ipv4Address source, 
            Ipv4Address destination,
            uint8_t protocol,
            Ptr<Ipv4Route> route)
{
  NS_LOG_FUNCTION (this << packet << source << destination << uint32_t(protocol) << route);

  Ipv4Header ipHeader;
  bool mayFragment = true;
  uint8_t ttl = m_defaultTtl;
  SocketIpTtlTag tag;
  bool found = packet->RemovePacketTag (tag);
  if (found)
    {
      ttl = tag.GetTtl ();
    }

  // Handle a few cases:
  // 1) packet is destined to limited broadcast address
  // 2) packet is destined to a subnet-directed broadcast address
  // 3) packet is not broadcast, and is passed in with a route entry
  // 4) packet is not broadcast, and is passed in with a route entry but route->GetGateway is not set (e.g., on-demand)
  // 5) packet is not broadcast, and route is NULL (e.g., a raw socket call, or ICMP)
  
  // 1) packet is destined to limited broadcast address
  if (destination.IsBroadcast ()) 
    {
      NS_LOG_LOGIC ("Ipv4L3Protocol::Send case 1:  limited broadcast");
      ttl = 1;
      ipHeader = BuildHeader (source, destination, protocol, packet->GetSize (), ttl, mayFragment);
      uint32_t ifaceIndex = 0;
      for (Ipv4InterfaceList::iterator ifaceIter = m_interfaces.begin ();
           ifaceIter != m_interfaces.end (); ifaceIter++, ifaceIndex++)
        {
          Ptr<Ipv4Interface> outInterface = *ifaceIter;
          Ptr<Packet> packetCopy = packet->Copy ();

          NS_ASSERT (packetCopy->GetSize () <= outInterface->GetDevice()->GetMtu ());
          packetCopy->AddHeader (ipHeader);
          m_txTrace (packetCopy, ifaceIndex);
          outInterface->Send (packetCopy, destination);
        }
      return;
    }

  // 2) check: packet is destined to a subnet-directed broadcast address
  uint32_t ifaceIndex = 0;
  for (Ipv4InterfaceList::iterator ifaceIter = m_interfaces.begin ();
    ifaceIter != m_interfaces.end (); ifaceIter++, ifaceIndex++)
    {
      Ptr<Ipv4Interface> outInterface = *ifaceIter;
      for (uint32_t j = 0; j < GetNAddresses (ifaceIndex); j++)
        {
          Ipv4InterfaceAddress ifAddr = GetAddress (ifaceIndex, j);
          NS_LOG_LOGIC ("Testing address " << ifAddr.GetLocal () << " with mask " << ifAddr.GetMask ());
          if (destination.IsSubnetDirectedBroadcast (ifAddr.GetMask ()) && 
              destination.CombineMask (ifAddr.GetMask ()) == ifAddr.GetLocal ().CombineMask (ifAddr.GetMask ())   )  
            {
              NS_LOG_LOGIC ("Ipv4L3Protocol::Send case 2:  subnet directed bcast to " << ifAddr.GetLocal ());
              ttl = 1;
              ipHeader = BuildHeader (source, destination, protocol, packet->GetSize (), ttl, mayFragment);
              Ptr<Packet> packetCopy = packet->Copy ();
              packetCopy->AddHeader (ipHeader);
              m_txTrace (packetCopy, ifaceIndex);
              outInterface->Send (packetCopy, destination);
              return;
            }
        }
    }

  // 3) packet is not broadcast, and is passed in with a route entry
  //    with a valid Ipv4Address as the gateway
  if (route && route->GetGateway () != Ipv4Address ())
    {
      NS_LOG_LOGIC ("Ipv4L3Protocol::Send case 3:  passed in with route");
      ipHeader = BuildHeader (source, destination, protocol, packet->GetSize (), ttl, mayFragment);
      SendRealOut (route, packet, ipHeader);
      return; 
    } 
  // 4) packet is not broadcast, and is passed in with a route entry but route->GetGateway is not set (e.g., on-demand)
  if (route && route->GetGateway () != Ipv4Address ())
    {
      // This could arise because the synchronous RouteOutput() call
      // returned to the transport protocol with a source address but
      // there was no next hop available yet (since a route may need
      // to be queried).  
      NS_FATAL_ERROR ("This case not yet implemented");
    }
  // 5) packet is not broadcast, and route is NULL (e.g., a raw socket call)
  NS_LOG_LOGIC ("Ipv4L3Protocol::Send case 4:  passed in with no route " << destination);
  Socket::SocketErrno errno_; 
  uint32_t oif = 0; // unused for now
  ipHeader = BuildHeader (source, destination, protocol, packet->GetSize (), ttl, mayFragment);
  Ptr<Ipv4Route> newRoute = m_routingProtocol->RouteOutput (packet, ipHeader, oif, errno_);
  if (newRoute)
    {
      SendRealOut (newRoute, packet, ipHeader);
    }
  else
    {
      NS_LOG_WARN ("No route to host.  Drop.");
      m_dropTrace (packet);
    }
}

// XXX when should we set ip_id?   check whether we are incrementing
// m_identification on packets that may later be dropped in this stack
// and whether that deviates from Linux
Ipv4Header
Ipv4L3Protocol::BuildHeader (
            Ipv4Address source, 
            Ipv4Address destination,
            uint8_t protocol,
            uint16_t payloadSize,
            uint8_t ttl,
            bool mayFragment)
{
  NS_LOG_FUNCTION_NOARGS ();
  Ipv4Header ipHeader;
  ipHeader.SetSource (source);
  ipHeader.SetDestination (destination);
  ipHeader.SetProtocol (protocol);
  ipHeader.SetPayloadSize (payloadSize);
  ipHeader.SetTtl (ttl);
  if (mayFragment == true)
    {
      ipHeader.SetMayFragment ();
      ipHeader.SetIdentification (m_identification);
      m_identification ++;
    }
  else
    {
      ipHeader.SetDontFragment ();
      // TBD:  set to zero here; will cause traces to change
      ipHeader.SetIdentification (m_identification);
      m_identification ++;
    }
  if (Node::ChecksumEnabled ())
    {
      ipHeader.EnableChecksum ();
    }
  return ipHeader;
}

void
Ipv4L3Protocol::SendRealOut (Ptr<Ipv4Route> route,
                             Ptr<Packet> packet,
                             Ipv4Header const &ipHeader)
{
  NS_LOG_FUNCTION (this << packet << &ipHeader);

  // We add a header regardless of whether there is a route, since 
  // we may want to drop trace
  packet->AddHeader (ipHeader);
  if (route == 0)
    {
      NS_LOG_WARN ("No route to host.  Drop.");
      m_dropTrace (packet);
      return;
    }

  Ptr<NetDevice> outDev = route->GetOutputDevice ();
  int32_t interface = GetInterfaceForDevice (outDev);
  NS_ASSERT (interface >= 0);
  Ptr<Ipv4Interface> outInterface = GetInterface (interface);
  NS_LOG_LOGIC ("Send via NetDevice ifIndex " << outDev->GetIfIndex () << " ipv4InterfaceIndex " << interface);

  NS_ASSERT (packet->GetSize () <= outInterface->GetDevice ()->GetMtu ());
  if (!route->GetGateway ().IsEqual (Ipv4Address ("0.0.0.0"))) 
    {
      if (outInterface->IsUp ())
        {
          NS_LOG_LOGIC ("Send to gateway " << route->GetGateway ());
          m_txTrace (packet, interface);
          outInterface->Send (packet, route->GetGateway ());
        }
      else
        {
          NS_LOG_LOGIC ("Dropping-- outgoing interface is down: " << route->GetGateway ());
          m_dropTrace (packet);
        }
    } 
  else 
    {
      if (outInterface->IsUp ())
        {
          NS_LOG_LOGIC ("Send to destination " << ipHeader.GetDestination ());
          m_txTrace (packet, interface);
          outInterface->Send (packet, ipHeader.GetDestination ());
        }
      else
        {
          NS_LOG_LOGIC ("Dropping-- outgoing interface is down: " << ipHeader.GetDestination ());
          m_dropTrace (packet);
        }
    }
}

// This function analogous to Linux ip_mr_forward()
void
Ipv4L3Protocol::IpMulticastForward (Ptr<Ipv4MulticastRoute> mrtentry, Ptr<const Packet> p, const Ipv4Header &header)
{
  NS_LOG_FUNCTION (mrtentry << p << header);
  NS_LOG_LOGIC ("Multicast forwarding logic for node: " << m_node->GetId ());
  // The output interfaces we could forward this onto are encoded
  // in the OutputTtl of the Ipv4MulticastRoute
  for (uint32_t i = 0; i < Ipv4MulticastRoute::MAX_INTERFACES; i++)
    {
      if (mrtentry->GetOutputTtl (i) < Ipv4MulticastRoute::MAX_TTL)
        {
          Ptr<Packet> packet = p->Copy ();
          Ipv4Header h = header;
          h.SetTtl (header.GetTtl () - 1);
          if (h.GetTtl () == 0)
            {
              NS_LOG_WARN ("TTL exceeded.  Drop.");
              m_dropTrace (packet);
              return;
            }
          NS_LOG_LOGIC ("Forward multicast via interface " << i);
          Ptr<Ipv4Route> rtentry = Create<Ipv4Route> ();
          rtentry->SetSource (h.GetSource ());
          rtentry->SetDestination (h.GetDestination ());
          rtentry->SetGateway (Ipv4Address::GetAny ());
          rtentry->SetOutputDevice (GetNetDevice (i));
          SendRealOut (rtentry, packet, h);
          return; 
        }
    }
}

// This function analogous to Linux ip_forward()
void
Ipv4L3Protocol::IpForward (Ptr<Ipv4Route> rtentry, Ptr<const Packet> p, const Ipv4Header &header)
{
  NS_LOG_FUNCTION (rtentry << p << header);
  NS_LOG_LOGIC ("Forwarding logic for node: " << m_node->GetId ());
  // Forwarding
  Ipv4Header ipHeader = header;
  Ptr<Packet> packet = p->Copy ();
  ipHeader.SetTtl (ipHeader.GetTtl () - 1);
  if (ipHeader.GetTtl () == 0)
    {
      // Do not reply to ICMP or to multicast/broadcast IP address 
      if (ipHeader.GetProtocol () != Icmpv4L4Protocol::PROT_NUMBER && 
        ipHeader.GetDestination ().IsBroadcast () == false && 
        ipHeader.GetDestination ().IsMulticast () == false)
        {
          Ptr<Icmpv4L4Protocol> icmp = GetIcmp ();
          icmp->SendTimeExceededTtl (ipHeader, packet);
        }
      NS_LOG_WARN ("TTL exceeded.  Drop.");
      m_dropTrace (packet);
      return;
    }
  SendRealOut (rtentry, packet, ipHeader);
}

void
Ipv4L3Protocol::LocalDeliver (Ptr<const Packet> packet, Ipv4Header const&ip, uint32_t iif)
{
  NS_LOG_FUNCTION (this << packet << &ip);
  Ptr<Packet> p = packet->Copy (); // need to pass a non-const packet up

  Ptr<Ipv4L4Protocol> protocol = GetProtocol (ip.GetProtocol ());
  if (protocol != 0)
    {
      // we need to make a copy in the unlikely event we hit the
      // RX_ENDPOINT_UNREACH codepath
      Ptr<Packet> copy = p->Copy ();
      enum Ipv4L4Protocol::RxStatus status = 
        protocol->Receive (p, ip.GetSource (), ip.GetDestination (), GetInterface (iif));
      switch (status) {
      case Ipv4L4Protocol::RX_OK:
        // fall through
      case Ipv4L4Protocol::RX_CSUM_FAILED:
        break;
      case Ipv4L4Protocol::RX_ENDPOINT_UNREACH:
        if (ip.GetDestination ().IsBroadcast () == true || 
          ip.GetDestination ().IsMulticast () == true)
          {
            break;  // Do not reply to broadcast or multicast
          }
        // Another case to suppress ICMP is a subnet-directed broadcast
        bool subnetDirected = false;
        for (uint32_t i = 0; i < GetNAddresses (iif); i++)
          {
            Ipv4InterfaceAddress addr = GetAddress (iif, i);
            if (addr.GetLocal ().CombineMask (addr.GetMask ()) == ip.GetDestination().CombineMask (addr.GetMask ()) &&
              ip.GetDestination ().IsSubnetDirectedBroadcast (addr.GetMask ()))
              {
                subnetDirected = true;
              }
          }
        if (subnetDirected == false)
          {
            GetIcmp ()->SendDestUnreachPort (ip, copy);
          }
      }
    }
}

bool
Ipv4L3Protocol::AddAddress (uint32_t i, Ipv4InterfaceAddress address)
{
  NS_LOG_FUNCTION (this << i << address);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  bool retVal = interface->AddAddress (address);
  if (m_routingProtocol != 0)
    {
      m_routingProtocol->NotifyAddAddress (i, address);
    }
  return retVal;
}

Ipv4InterfaceAddress 
Ipv4L3Protocol::GetAddress (uint32_t interfaceIndex, uint32_t addressIndex) const
{
  NS_LOG_FUNCTION (this << interfaceIndex << addressIndex);
  Ptr<Ipv4Interface> interface = GetInterface (interfaceIndex);
  return interface->GetAddress (addressIndex);
}

uint32_t 
Ipv4L3Protocol::GetNAddresses (uint32_t interface) const
{
  NS_LOG_FUNCTION (this << interface);
  Ptr<Ipv4Interface> iface = GetInterface (interface);
  return iface->GetNAddresses ();
}

bool
Ipv4L3Protocol::RemoveAddress (uint32_t i, uint32_t addressIndex)
{
  NS_LOG_FUNCTION (this << i << addressIndex);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  Ipv4InterfaceAddress address = interface->RemoveAddress (addressIndex);
  if (address != Ipv4InterfaceAddress ())
    {
      if (m_routingProtocol != 0)
        {
          m_routingProtocol->NotifyRemoveAddress (i, address);
        }
      return true;
    }
  return false;
}

void 
Ipv4L3Protocol::SetMetric (uint32_t i, uint16_t metric)
{
  NS_LOG_FUNCTION (i << metric);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  interface->SetMetric (metric);
}

uint16_t
Ipv4L3Protocol::GetMetric (uint32_t i) const
{
  NS_LOG_FUNCTION (i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  return interface->GetMetric ();
}

uint16_t 
Ipv4L3Protocol::GetMtu (uint32_t i) const
{
  NS_LOG_FUNCTION (this << i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  return interface->GetDevice ()->GetMtu ();
}

bool 
Ipv4L3Protocol::IsUp (uint32_t i) const
{
  NS_LOG_FUNCTION (this << i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  return interface->IsUp ();
}

void 
Ipv4L3Protocol::SetUp (uint32_t i)
{
  NS_LOG_FUNCTION (this << i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  interface->SetUp ();

  if (m_routingProtocol != 0)
    {
      m_routingProtocol->NotifyInterfaceUp (i);
    }
}

void 
Ipv4L3Protocol::SetDown (uint32_t ifaceIndex)
{
  NS_LOG_FUNCTION (this << ifaceIndex);
  Ptr<Ipv4Interface> interface = GetInterface (ifaceIndex);
  interface->SetDown ();

  if (m_routingProtocol != 0)
    {
      m_routingProtocol->NotifyInterfaceDown (ifaceIndex);
    }
}

bool 
Ipv4L3Protocol::IsForwarding (uint32_t i) const
{
  NS_LOG_FUNCTION (this << i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  NS_LOG_LOGIC ("Forwarding state: " << interface->IsForwarding ());
  return interface->IsForwarding ();
}

void 
Ipv4L3Protocol::SetForwarding (uint32_t i, bool val)
{
  NS_LOG_FUNCTION (this << i);
  Ptr<Ipv4Interface> interface = GetInterface (i);
  interface->SetForwarding (val);
}

Ptr<NetDevice>
Ipv4L3Protocol::GetNetDevice (uint32_t i)
{
  NS_LOG_FUNCTION (this << i);
  return GetInterface (i)-> GetDevice ();
}

void 
Ipv4L3Protocol::SetIpForward (bool forward) 
{
  NS_LOG_FUNCTION (this << forward);
  m_ipForward = forward;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); i != m_interfaces.end (); i++)
    {
      (*i)->SetForwarding (forward);
    }
}

bool 
Ipv4L3Protocol::GetIpForward (void) const
{
  return m_ipForward;
}

void
Ipv4L3Protocol::RouteInputError (Ptr<const Packet> p, const Ipv4Header & ipHeader, Socket::SocketErrno sockErrno)
{
  NS_LOG_FUNCTION (this << p << ipHeader << sockErrno);
  NS_LOG_LOGIC ("Route input failure-- dropping packet to " << ipHeader << " with errno " << sockErrno); 
  m_dropTrace (p);
}

///////////////////////////Naveed///////////////////////////
void
Ipv4L3Protocol::CreateApIpv4List (void)
{
  Ipv4Address ipv4Address ("10.1.3.1");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("10.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("11.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);  
  ipv4Address.Set ("12.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("13.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("14.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);  
  ipv4Address.Set ("15.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("16.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
  ipv4Address.Set ("17.1.3.2");
  m_apIpv4Address.push_back (ipv4Address);
}

void
Ipv4L3Protocol::ConnectToCatchAssocEvent (void)
{
/*  Ptr<NetDevice> pNetDevice;
  Ptr<WifiNetDevice> pWifiNetDevice;
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); i != m_interfaces.end (); i++)
    {
      pNetDevice = (*i)->GetDevice ();
      pWifiNetDevice = (Ptr<WifiNetDevice>) pNetDevice;
    }
*/

  std::ostringstream oss;
//  std::cout << m_node->GetId () << "\n";
  oss << "/NodeList/" << m_node->GetId () << "/DeviceList/*/Mac/$ns3::NqapWifiMac/AssocNotif";
  Config::Connect (oss.str (), MakeCallback (&Ipv4L3Protocol::ReceiveAssocNotif, this));

  std::ostringstream oss2;  
  oss2 << "/NodeList/" << m_node->GetId () << "/DeviceList/*/Mac/$ns3::NqapWifiMac/DisAssocNotif";
  Config::Connect (oss2.str (), MakeCallback (&Ipv4L3Protocol::ReceiveDisAssocNotif, this));
}

void
Ipv4L3Protocol::ReceiveAssocNotif (std::string context, Mac48Address staAddress, Mac48Address apAddress)
{
/*
  Ptr<Packet> packet = Create<Packet> ();
  
  std::cout << "Associated Event is received from a lower layer: " << staAddress << " associated with " << apAddress << "\n";
  
  Ptr<ArpL3Protocol> arp = m_node->GetObject<ArpL3Protocol> ();
  
  Ipv4Address ipv4AddressSta = arp->SearchForIpv4Address (staAddress);
  Ipv4Address ipv4AddressAP = arp->SearchForIpv4Address (apAddress);
//  std::cout << "addresses returned are " << ipv4AddressSta << " & " << ipv4AddressAP << "\n";

  if (ipv4AddressSta == "127.0.0.1")
    {
      std::cout << "Station Address " << ipv4AddressSta << " could not be translated\n";
      return;
    }

  if (ipv4AddressAP == "127.0.0.1")
    {
      std::cout << "AP Address " << ipv4AddressAP << " could not be translated\n";
      return;
    }

  uint32_t ifIndex = GetInterfaceForAddress (ipv4AddressAP);
  RemoveRoutingEntry (ipv4AddressSta, ifIndex);

  m_hNotif.AddOrModifyDestInfo (ipv4AddressSta, ipv4AddressAP, ipv4AddressSta);
  m_hNotif.SendNotificationToAPs (ASS_NOTIF, 0, ipv4AddressAP, ipv4AddressSta, ipv4AddressSta, packet);
 
  ApIpv4AddressList::iterator newIterator;
  Ipv4Address destination;

  
//  std::cout << "ReceiveAssocNotif: Adding route for address " << ipv4AddressSta << "\n";
  Ptr<Ipv4ListRouting> ipv4LRP;
  ipv4LRP = DynamicCast<Ipv4ListRouting> (m_routingProtocol);
  int16_t priority1 = 0;
  Ptr<Ipv4StaticRouting> ipv4Static =  DynamicCast<Ipv4StaticRouting> (ipv4LRP->GetRoutingProtocol (0, priority1));
  
  ipv4Static->AddHostRouteTo (ipv4AddressSta, ifIndex);

  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
    {
      Ptr<Packet> p = packet->Copy ();
      destination = *newIterator;

      if (CheckForLocalInterface (destination))
        {
    continue;
  }
      else
        {
          Send (p, ipv4AddressAP, destination, 17, NULL);
  }
    }
  m_hNotif.RemoveDestBufferInfo (ipv4AddressSta);
    
  Ptr<Packet> packet2 = Create<Packet> ();
  m_hNotif.SendNotificationToAPs (FETCH_FRAMES_NOTIF, 0, ipv4AddressAP, ipv4AddressSta, ipv4AddressSta, packet2);
  
  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
    {
      Ptr<Packet> p = packet2->Copy ();
      destination = *newIterator;

      if (CheckForLocalInterface (destination))
        {
    continue;
  }
      else
        {
          Send (p, ipv4AddressAP, destination, 17, NULL);
  }
    }
    
  int numStoredFrames = CountNumberOfStoredFrames(ipv4AddressSta);

  std::cout << "ReceiveAssocNotif (), Number of Stored Frames for dest " << ipv4AddressSta << " is = " << numStoredFrames << "\n";
  std::cout << "ReceiveAssocNotif (), Total Stored Frames are = " << buffer.m_listSize << "\n";

  if (numStoredFrames > 0)
    {
      if (m_txTimerEvent.IsRunning ())
          m_txTimerEvent.Cancel ();
      SendStoreFramesToDest();
      m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
    }
*/
}

void
Ipv4L3Protocol::ReceiveDisAssocNotif (std::string context, Mac48Address staAddress, Mac48Address apAddress)
{
  Ptr<Packet> packet = Create<Packet> ();

  Ptr<ArpL3Protocol> arp = m_node->GetObject<ArpL3Protocol> ();
//  Mac48Address IPAddress ("00:00:00:00:00:15");

  Ipv4Address ipv4AddressSta = arp->SearchForIpv4Address (staAddress);
  Ipv4Address ipv4AddressAP = arp->SearchForIpv4Address (apAddress);

  std::cout << Simulator::Now () << " ---> Disassociated Event is received from a lower layer: " << ipv4AddressSta << " disassociated from " << ipv4AddressAP << "\n";
    
  if (ipv4AddressSta == "127.0.0.1")
    {
#ifdef DEBUG
      std::cout << "Loopback Address is returned\n";
#endif
    }

  if (ipv4AddressAP == "127.0.0.1")
    {
#ifdef DEBUG
      std::cout << "Loopback Address is returned\n";
#endif
    }

  ///////// TO BE ADDRESSED --- alternateAddress --- //////////
  Ipv4Address alternateAddress[2];
  uint16_t len;
  m_hNotif.RemoveDestInfo (ipv4AddressSta, alternateAddress, &len);
  
  Ipv4Address address[3];
  address[0] = ipv4AddressSta;
  address[1] = alternateAddress[0];
  address[2] = alternateAddress[1];
  
  m_hNotif.CheckAndAddDestBufferAP (address, ipv4AddressAP, len+1);
  
  if (buffer.m_listSize == 0)
    {
      if (m_txTimerEvent.IsRunning ())
          m_txTimerEvent.Cancel ();
    }
    
  uint32_t ifIndex = GetInterfaceForAddress (ipv4AddressAP);
  
  RemoveRoutingEntry (ipv4AddressSta, ifIndex);
  
  if (len >= 2)
    {
      RemoveRoutingEntry (alternateAddress[0], ifIndex);
    }
  if (len >= 3)
    {
      RemoveRoutingEntry (alternateAddress[1], ifIndex);
    }
  
//  Ipv4Address destAddress;
  RemoveRouteWithGWAsDest (ipv4AddressSta, ipv4AddressAP);

/*  
  if (destAddress != "102.102.102.102")
    {
      m_hNotif.RemoveDestInfo (destAddress, &alternateAddress);
    }


  m_hNotif.SendNotificationToAPs (DISASS_NOTIF, 0, ipv4AddressSta, ipv4AddressAP, ipv4AddressSta, packet);

  ApIpv4AddressList::iterator newIterator;
  Ipv4Address destination;
  
  uint32_t ifIndex = GetInterfaceForAddress (ipv4AddressAP);
  RemoveRoutingEntry (ipv4AddressSta, ifIndex);
  
  for (newIterator = m_apIpv4Address.begin (); newIterator != m_apIpv4Address.end (); newIterator++)
    {
      Ptr<Packet> p = packet->Copy ();
      destination = *newIterator;

      if (CheckForLocalInterface (destination))
        {
    continue;
  }
      else
        {
          Send (p, ipv4AddressAP, destination, 17);
  }
    }
*/
}


bool
Ipv4L3Protocol::CheckForLocalInterface (Ipv4Address destination)
{
  bool found = false;
  Ipv4Address ipAddr;
  
  for (Ipv4InterfaceList::const_iterator i = m_interfaces.begin (); i != m_interfaces.end (); i++)
    {
      for (uint32_t j = 0; j < (*i)->GetNAddresses (); j++)
        {
          if ((*i)->GetAddress (j).GetLocal () == destination)
            {
        found = true;
              break;
            }
        }
      if (found)
        {
    break;
  }
    }
  return found;
}


int
Ipv4L3Protocol::CountNumberOfStoredFrames(Ipv4Address destAddress)
{
  int counter = 0;
  
  std::list<BufferedFramesSt>::iterator newIterator;

  for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); newIterator++)
    {
      if (destAddress == newIterator->dst)
        {
          counter++;
        }      
    }
  return counter;
}

HandleNotificationsClass 
Ipv4L3Protocol::GetHandleNotificationClass (void)
{
  return m_hNotif;
}

int
Ipv4L3Protocol::NeedToBufferPacket (Ipv4Address destination)
{
  int status = 0;
  status = m_hNotif.NeedToBufferPacket (destination, m_interfaces);
  return status;
}

void
Ipv4L3Protocol::MeDeHaBufferPacketLocally (Ptr<const Packet> packet, Ipv4Header ipHeader)
{
#ifdef BUFFER
//  Packet pkt = *packet;
  Ptr<Packet> pkt = packet->Copy ();
  
  MedehaIntraManetNotifHeader hdr;
  
  if (ipHeader.GetProtocol () == 18)
    {
      std::cout << "Check Point 10\n";
      pkt->RemoveHeader (hdr);
    }

  UdpHeader udpHdr;
  pkt->RemoveHeader (udpHdr);
  SeqNumHeader seqHdr;
  pkt->RemoveHeader (seqHdr);
  int seqNum = seqHdr.GetSeqNum ();
  int priority = seqHdr.GetPriority ();
  int copies = seqHdr.GetNumOfCopies ();

  //#ifdef DEBUG
  std::cout << "Ipv4L3Protocol::MeDeHaBufferPacketLocally, Packet with Seq Num " << seqNum << " buffered\n";
  //#endif
  
  pkt->AddHeader (seqHdr);
  pkt->AddHeader (udpHdr);
  
  VectorSummery msgVectorSummery;
  msgVectorSummery.srcAddress = ipHeader.GetSource ();
  msgVectorSummery.dstAddress = ipHeader.GetDestination ();
  msgVectorSummery.seqNumber = seqNum;
  m_msgVectorSummeryList.push_back (msgVectorSummery);
  

  BufferedFramesSt frameForBuffer;
  frameForBuffer.packet = pkt;
  frameForBuffer.src = ipHeader.GetSource ();
  
  if (ipHeader.GetProtocol () == 18)
    {
      std::cout << "Check Point 11\n";
      frameForBuffer.dst = hdr.GetDstIPAddress ();
      frameForBuffer.protocol = hdr.GetProtocol ();
    }
  else
    {
      frameForBuffer.dst = ipHeader.GetDestination ();
      frameForBuffer.protocol = ipHeader.GetProtocol ();
    }
  frameForBuffer.priority = priority;
  frameForBuffer.ttl = 20;
  
  if (copies > 0 && copies < MAX_NUM_OF_COPIES)
    {
      buffer.AddElementToList(frameForBuffer);
    }
#ifdef DEBUG
  std::cout << "Ipv4L3Protocol::MeDeHaBufferPacketLocally, buffering data for " << ipHeader.GetDestination () << "\n";
#endif
#endif
}


void 
Ipv4L3Protocol::RemoveRoutingEntry (Ipv4Address destAddr, uint32_t ifIndex)
{
  Ptr<Ipv4ListRouting> ipv4LRP;
  ipv4LRP = DynamicCast<Ipv4ListRouting> (m_routingProtocol);
  int16_t priority1 = 0;
  Ptr<Ipv4StaticRouting> ipv4Static =  DynamicCast<Ipv4StaticRouting> (ipv4LRP->GetRoutingProtocol (0, priority1));

  for (uint32_t i = 0; i < ipv4Static->GetNRoutes (); i++)
    {
      Ipv4RoutingTableEntry route = ipv4Static->GetRoute (i);
      if (route.GetDest () == destAddr)
        {
#ifdef DEBUG
    std::cout << "Ipv4L3Protocol::RemoveRoutingEntry, Removing Route for " << destAddr << ", Gateway is " << route.GetGateway () << ", and Interface is " << route.GetInterface () << "\n";
#endif
          ipv4Static->RemoveRoute (i);
          break;
        }
    }    
}

void
Ipv4L3Protocol::RemoveRouteWithGWAsDest (Ipv4Address gatewayAddress, Ipv4Address ipv4AddressAP)
{
  Ipv4Address destAddress;
  Ptr<Ipv4ListRouting> ipv4LRP;
  ipv4LRP = DynamicCast<Ipv4ListRouting> (m_routingProtocol);
  int16_t priority1 = 0;
  Ptr<Ipv4StaticRouting> ipv4Static =  DynamicCast<Ipv4StaticRouting> (ipv4LRP->GetRoutingProtocol (0, priority1));

  for (uint32_t i = 0; i < ipv4Static->GetNRoutes (); i++)
    {
      Ipv4RoutingTableEntry route = ipv4Static->GetRoute (i);
      if (route.GetGateway () == gatewayAddress)
        {
    destAddress = route.GetDest ();
#ifdef DEBUG
    std::cout << "Ipv4L3Protocol::RemoveRouteWithGWAsDest, Removing Route for " << route.GetDest () << ", Gateway is " << route.GetGateway () << ", and Interface is " << route.GetInterface () << "\n";
#endif
          ipv4Static->RemoveRoute (i);

    Ipv4Address alternateAddress[2];
    uint16_t len;
          int status = m_hNotif.RemoveDestInfo (destAddress, alternateAddress, &len);

          if (status == DEST_INFO_ENTRY_REMOVED)
      {
        Ipv4Address address[3];
              address[0] = destAddress;
              if (len >= 1)
                  address[1] = alternateAddress[0];
              if (len >= 2)
                  address[2] = alternateAddress[1];

              m_hNotif.CheckAndAddDestBufferAP (address, ipv4AddressAP, len+1);
      }
          continue;
        }
    }
}

void
Ipv4L3Protocol::HandleSendTimer (void)
{
//  std::cout << "********* global store counter for ap (" << this->GetAddress () << ") = " << globalStoreCounter << " *********\n";

  uint16_t i=0;
  if (buffer.m_listSize > 0)
    {
      NS_LOG_INFO("Timer Expires, Sending stored frames");
      i = SendStoreFramesToDest();
    }
  if (buffer.m_listSize > 0 && i > 0)
    {
      m_txTimerEvent = Simulator::Schedule (m_txTimer, &Ipv4L3Protocol::HandleSendTimer, this); // reschedule the timer for transmission of buffered frames
    }
  else
    {
      m_txTimerEvent.Cancel ();
    }
}

uint16_t
Ipv4L3Protocol::SendStoreFramesToDest (void)
{
  Ptr<Packet> packet;
  int counter = 0;
  int i = 0;
  Ipv4Address destAddress, alternateAddress01, alternateAddress02;
  
#ifdef DEBUG
  std::cout << "Ipv4L3Protocol::SendStoreFramesToDest is called. Size is = " << buffer.GetBufferSize () << "\n";
#endif

  bool flag = 0;
  std::list<DestInfoSt>::iterator destInfoIterator;
  Ipv4Address apAddress;
  for (destInfoIterator = m_hNotif.destInfoList.begin(), i=0; ; )
    {
      destAddress = destInfoIterator->destNodeAddress;
      apAddress = destInfoIterator->assAPAddress;
      alternateAddress01 = destInfoIterator->destNodeAltAddress01;
      alternateAddress02 = destInfoIterator->destNodeAltAddress02;
//    std::cout << "SendStoreFramesToDest(): dest address = " << destAddress << " my address = " << myAddress << endl;    

      int count01 = CountNumberOfStoredFrames (destAddress);
      int count02 = CountNumberOfStoredFrames (alternateAddress01);
      int count03 = CountNumberOfStoredFrames (alternateAddress02);
      
//    std::cout << "SendStoreFramesToDest(): Count for dest " << destAddress << " is " << count << endl;

      if (count01 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;    // link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (destAddress == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
      std::cout << "SendStoreFramesToDest(): Sending message to " << destAddress << "\n";
#endif
      Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
      newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
      buffer.SetMemoryState ();
//                  i++;
                  break;
                }
              else
                {
//                  EV << "Dest Address = " << destAddress << " doesn't match. Checking next entry\n";
                  counter++;
                  newIterator++;
                }                   
            }
        }
  
      if (count02 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;    // link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (alternateAddress01 == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
      std::cout << "SendStoreFramesToDest(): Sending message to " << alternateAddress01 << "\n";
#endif
      Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
      newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
      buffer.SetMemoryState ();
//                  i++;
                  break;
                }
              else
                {
//                  EV << "Dest Address = " << alternateAddress << " doesn't match. Checking next entry\n";
                  counter++;
                  newIterator++;
                }
            }
        }
  
      if (count03 > 0)
        {
          std::list<BufferedFramesSt>::iterator newIterator;    // link list iterator
          for (newIterator = buffer.m_pktList.begin(); newIterator != buffer.m_pktList.end(); )
            {
              packet = (newIterator->packet);

              if (alternateAddress02 == newIterator->dst)
                {
                  flag = 1;
#ifdef DEBUG
      std::cout << "SendStoreFramesToDest(): Sending message to " << alternateAddress02 << "\n";
#endif
      Send (packet, newIterator->src, newIterator->dst, newIterator->protocol, NULL);
      newIterator = buffer.m_pktList.erase(newIterator);
                  (buffer.m_listSize)--;
      buffer.SetMemoryState ();
//                  i++;
                  break;
                }
              else
                {
//                  EV << "Dest Address = " << alternateAddress << " doesn't match. Checking next entry\n";
                  counter++;
                  newIterator++;
                }
            }
        }
  
      if (flag)
        {
    i++;
  }

      destInfoIterator++;   
      if (i == (MAX_NUM_FRAMES_TO_SEND) || buffer.m_pktList.size() == 0)
        {       
          break;
        }
      if (destInfoIterator == m_hNotif.destInfoList.end())
        {
          if (flag)
            {
              destInfoIterator = m_hNotif.destInfoList.begin();
              flag = 0;
            }
          else
            {
              break;
            }
        }
    }
  return i;
}


void
Ipv4L3Protocol::RemoveManetRoutes (Ipv4Address gateway, Ipv4Address apAddress)
{
  std::cout << "Ipv4L3Protocol::RemoveManetRoutes (), At " << apAddress << ", gateway = " << gateway << "\n";
  Ipv4Address destAddr, gatewayAddr;
  uint16_t manetConnection;

  std::list<DestInfoSt>::iterator destInfoIterator;
  for (destInfoIterator = m_hNotif.destInfoList.begin(); destInfoIterator != m_hNotif.destInfoList.end(); )
    {
      destAddr = destInfoIterator->destNodeAddress;
      gatewayAddr = destInfoIterator->assAPAddress;
      manetConnection = destInfoIterator->m_manetConnection;
/*      
      std::cout << "Dest: " << destAddr << std::endl;
      std::cout << "GW: " << gatewayAddr << std::endl;
      std::cout << "Manet: " << manetConnection << std::endl;
      std::cout << "Alt01: " << destInfoIterator->destNodeAltAddress01 << std::endl;
      std::cout << "Alt02: " << destInfoIterator->destNodeAltAddress02 << std::endl << "\n";
*/      
      if (manetConnection)
        {
    if (gatewayAddr == gateway)
      {
        Ipv4Address alternateAddress;
        destInfoIterator = m_hNotif.destInfoList.erase (destInfoIterator);
        std::cout << "Ipv4L3Protocol::RemoveManetRoutes (), Removing MANET node " << destAddr << "\n";
        RemoveRoutingEntry (destAddr, 0);
        m_hNotif.CheckAndAddDestBufferAP (&destAddr, apAddress, 1);
      }
    else
      {
        destInfoIterator++;
      }
  }
      else
        {
    destInfoIterator++;
  }
    }      
}

///////////////////////////Naveed///////////////////////////
}//namespace ns3
