/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 INRIA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Rao Naveed Bin Rais <Naveed-Bin.Rais@inria.fr>
 */

#include "medeha-ap-buffer.h"
#include "ns3/log.h"

NS_LOG_COMPONENT_DEFINE ("MedehaApBuffer");

namespace ns3 {

BufferClass::BufferClass()
{
  m_listSize = 0;
  m_listBufferLimit = 15000;
  m_lowPriorityLossCounter = 0;
  m_highPriorityLossCounter = 0;
  m_ttlExpiredCounter = 0;
  m_state = MEMORY_EMPTY;
}

void 
BufferClass::SetBufferLimit (uint32_t size)
{
  m_listBufferLimit = size;
}

uint32_t
BufferClass::GetBufferLimit (void)
{
  return m_listBufferLimit;
}

uint32_t
BufferClass::GetBufferSize (void)
{
  return m_listSize;
}

void 
BufferClass::AddElementToList(BufferedFramesSt frameToAdd)
{
  BufferedFramesSt thisP;		// temporary variable to keep an entry of a buffered element

  // Check if the buffer has any space left to save this element
  if (m_listSize < m_listBufferLimit)
    {// if space is available
      // save the element in the buffer
      // packet should be added at the back of the list
      m_pktList.push_back(frameToAdd);
      NS_LOG_INFO("BufferClass::Adding Element to List " << m_listSize);
//      std::cout << "BufferClass::Adding Element to List " << m_listSize << "\n";
      m_listSize++;
      
      SetMemoryState ();
    }
  else if (IsPacketTagged(frameToAdd))
    {// if no space is available in the buffer and coming frame is of high priority
      BufferedFramesList::iterator newIterator;		// link list iterator
      bool removeFlag = false;	// flag to keep track if an entry is removed from the buffer

      // for loop to iterate through the buffer list
      // we don't increment iterator element in foor loop
      // this is because the iterator is automatically advanced by one if we remove an element
      for (newIterator = m_pktList.begin(); newIterator != m_pktList.end(); )
      {
        thisP = *newIterator;	// this iteration of the buffered element
      
        // check if this element is tagged (prioritized)
        if (!IsPacketTagged(thisP))
          {// if this element is of low priority
            // we can remove this element from the buffer to accomodate the newly arrived frame
            std::cout << "BufferClass::AddElementToList, Dropping Untagged Element\n";
            m_lowPriorityLossCounter++;		// another low priority frame is dropped

            newIterator = m_pktList.erase(newIterator);	// erase oldest low priority element from the buffer
            m_listSize--;		// list size is decremented
            removeFlag = true;	// an entry is removed, set flag to true

            std::cout << "BufferClass::AddElementToList, Adding Packet after removing one\n";
            // add new element to buffer now
            m_pktList.push_back(frameToAdd); // packet should be added at the back of the list
            m_listSize++;		// increment list size
	    
	    SetMemoryState ();

            break;
          }
        else
          {// if current element is of high priority
            newIterator++;		// go to next iterator to see if it is of low priority
          }
      }
 
      if (!removeFlag)
        {// if no element is removed
           // buffer is still full and all elements are of high priority
           // Remove oldest entered high pririty element
           std::cout << "BufferClass::AddElementToList, Dropping Oldest Element\n";
           m_highPriorityLossCounter++;			// another high priority frame is dropped

           BufferedFramesList::iterator newIterator;		// link list iterator
           newIterator = m_pktList.begin();	// go to the start of the list to remove one element
           newIterator = m_pktList.erase(newIterator);	// remove the oldest high priority element
           m_listSize--;		// decrement list size by one
      
          // double check if the buffer has any space
          if (m_listSize < m_listBufferLimit)
            {
              // add the new element to the end of the list 
              std::cout << "BufferClass::AddElementToList, Adding Packet after removing two\n";
              m_pktList.push_back(frameToAdd); // packet should be added at the back of the list rather 
              m_listSize++;	// increment list size by one
	      
	      SetMemoryState ();
            }
        }
    }
  else
    {// new frame is of low priority and the buffer is full with high priority frames
      BufferedFramesList::iterator newIterator;		// link list iterator
      bool removeFlag = false;	// flag to keep track if an entry is removed from the buffer

      // for loop to iterate through the buffer list
      // we don't increment iterator element in foor loop
      // this is because the iterator is automatically advanced by one if we remove an element
      for (newIterator = m_pktList.begin(); newIterator != m_pktList.end(); )
        {
          thisP = *newIterator;	// this iteration of the buffered element
      
          // check if this element is tagged (prioritized)
          if (!IsPacketTagged(thisP))
            {// if this element is of low priority
              // we can remove this element from the buffer to accomodate the newly arrived frame
              std::cout << "BufferClass::AddElementToList, Dropping Untagged Element\n";
              m_lowPriorityLossCounter++;		// another low priority frame is dropped

              newIterator = m_pktList.erase(newIterator);	// erase oldest low priority element from the buffer
              m_listSize--;		// list size is decremented
              removeFlag = true;	// an entry is removed, set flag to true

              std::cout << "BufferClass::AddElementToList, Adding Packet after removing one\n";
              // add new element to buffer now
              m_pktList.push_back(frameToAdd); // packet should be added at the back of the list
              m_listSize++;		// increment list size
	      
	      SetMemoryState ();

              break;
            }
          else
            {// if current element is of high priority
              newIterator++;		// go to next iterator to see if it is of low priority
            }
        }

      if (!removeFlag)
        {// all elements in the buffer are of high priority, we discard the newly arrived frame
          std::cout << "BufferClass::AddElementToList, Discarding the packet, Buffer is full\n";
          m_lowPriorityLossCounter++;				// another low priority frame is dropped
        }
    }
}


int 
BufferClass::IsPacketTagged(BufferedFramesSt frameToBuffer)
{
  // check the priority field of the frame
  if (frameToBuffer.priority == 1)
    {
      // the frame is of high priority
      return true;
    }
  else
    {
      // the frame is of low priority
      return false;
    }
}

void 
BufferClass::SetMemoryState (void)
{
  if (m_listSize > 0 &&
      m_listSize < m_listBufferLimit*0.7)
    {
      m_state = MEMORY_AVAILABLE;
    }
  else if (m_listSize >= m_listBufferLimit*0.7 &&
           m_listSize < m_listBufferLimit*0.9)
    {
      m_state = MEMORY_APPROACHING_CRITICAL;
    }
  else if (m_listSize >= m_listBufferLimit*0.9 &&
           m_listSize < m_listBufferLimit)
    {
      m_state = MEMORY_CRITICAL;
    }
  else if (m_listSize == m_listBufferLimit)
    {
      m_state = MEMORY_FULL;
    }
}

BufferClass::BufferState
BufferClass::GetMemoryState (void)
{
  return m_state;
}

/*
void BufferClass::DecrementTTLValueOfElements(void)
{
	printf("BufferClass::DecrementTTLValueOfElements, Decrementing TTL for Packets\n");
	
	std::list<Ieee80211DataFrame*>::iterator newIterator;		// link list iterator
	Ieee80211DataFrame* thisP;	

	hdr_udpbuffer* hp;
	hdr_ip* ip;

	for (newIterator = m_pktList.begin(); newIterator != m_pktList.end(); newIterator++)
	{
		thisP = (Ieee80211DataFrame*) *newIterator;		

		hp = hdr_udpbuffer::access(thisP);
		ip = hdr_ip::access(thisP);
		hp->ttl_--;
		
		if (hp->ttl_ == 0)
		{
			printf("BufferClass::DecrementTTLValueOfElements, TTL Expired, Dropping Packet\n");
//			CheckAndAddEntryInDropList(ip->daddr(), TTL_EXP_DROP);
			m_ttlExpiredCounter++;
			newIterator = m_pktList.erase(newIterator);
			m_listSize--;
		}
	}

}


void BufferClass::CheckAndAddEntryInDropList(nsaddr_t destAddress, int type)
{
//	list<dropPacketsInfo_t*>::iterator dpiIterator;		// link list iterator
	int isEntryExist = FALSE;
				
//	for (dpiIterator = ptrDropPacketsInfo_.begin(); dpiIterator != ptrDropPacketsInfo_.end(); dpiIterator++)
	for (int i=0; i<sizeDPI_; i++)
	{
		//if ((*dpiIterator)->nodeAddress_ == destAddress)
		if (dropPacketsInfo_[i].nodeAddress_ == destAddress)
		{
			isEntryExist = TRUE;
			switch (type)
			{
				case LP_PKT_DROP:
				{
					//((*dpiIterator)->m_listBufferLimitm_lowPriorityLossCounter)++;
					(dropPacketsInfo_[i].m_listBufferLimitm_lowPriorityLossCounter)++;
					break;
				}
				case HP_PKT_DROP:
				{
					//((*dpiIterator)->m_highPriorityLossCounter)++;
					(dropPacketsInfo_[i].m_highPriorityLossCounter)++;
					break;
				}
				case TTL_EXP_DROP:
				{
					//((*dpiIterator)->m_ttlExpiredCounter)++;
					(dropPacketsInfo_[i].m_ttlExpiredCounter)++;
					break;
				}
				default:
				{
					printf("Wrong TYPE of Entry\n");
					break;
				}
			}
		}
	}
	
	if (!isEntryExist)
	{
		//struct dropPacketsInfo_t* temp = new (struct dropPacketsInfo_t);

		// Set the fields of the structure to add the entry in the list
		//temp->nodeAddress_ = destAddress;	// Set the mobile node address
		
		dropPacketsInfo_[sizeDPI_].nodeAddress_ = destAddress;
		
		switch (type)
		{
			case LP_PKT_DROP:
			{				
				dropPacketsInfo_[sizeDPI_].m_listBufferLimitm_lowPriorityLossCounter = 1;
				dropPacketsInfo_[sizeDPI_].m_highPriorityLossCounter = 0;
				dropPacketsInfo_[sizeDPI_].m_ttlExpiredCounter = 0;
				break;
			}
			case HP_PKT_DROP:
			{
				dropPacketsInfo_[sizeDPI_].m_listBufferLimitm_lowPriorityLossCounter = 0;
				dropPacketsInfo_[sizeDPI_].m_highPriorityLossCounter = 1;
				dropPacketsInfo_[sizeDPI_].m_ttlExpiredCounter = 0;
				break;
			}
			case TTL_EXP_DROP:
			{
				dropPacketsInfo_[sizeDPI_].m_listBufferLimitm_lowPriorityLossCounter = 0;
				dropPacketsInfo_[sizeDPI_].m_highPriorityLossCounter = 0;
				dropPacketsInfo_[sizeDPI_].m_ttlExpiredCounter = 1;
				break;
			}
			default:
			{
				printf("Wrong TYPE of Entry\n");
				break;
			}
		}
		
		//ptrDropPacketsInfo_.push_front(temp);
		sizeDPI_++;
		//delete temp;
	}
}
*/

} // ns3 namespace

